"""iMVP helper functions.

"""

def load_sequences_from_fasta(fn):
    """This function is used for load sequences from a FASTA file into a pandas DataFrame.
    
    Parameters
    ----------
    fn: str
        The file to load.
        
    Returns
    -------
    pd.DataFrame
    """
    import pandas as pd
    from Bio import SeqIO
    N = 0
    data = {"seq": {}, "site": {}}
    for seq in SeqIO.parse(fn, "fasta"):
        data["seq"][N] = str(seq.seq)
        data["site"][N] = seq.id
        N += 1
    return pd.DataFrame(data)

def extract_fasta_and_draw_motifs(prefix, df, cluster_col="Cluster", filter=None, motif_column="seq", draw_logos=True):
    """This function is used for quick extraction of sequences strored in a DataFrame into a FASTA file and then draw the motif logos with Weblogo.
   
    Parameters
    ----------
    prefix: str
        The name of output path, required.
    df: pd.DataFrame
        A DataFrame containing the sequences used, required.
    cluster_col: str or tuple
        The column name for the clusters, default="Cluster".
    filter: boolean
        The column name used for filtering results, where only TRUE values will be used, default=None (not applied).
    motif_column: str or tuple
        The column that cotaining the motif sequences, default="motif_F10".
    draw_logos: boolean
        If use Weblogo to draw logos, default=True.
    
    Returns
    -------
    None
    """
    import os
    print("===============  {} ===============".format(prefix))
    if os.path.isdir("{}".format(prefix)) == False:
        os.mkdir("./{}".format(prefix))
        os.mkdir("./{}/fasta".format(prefix))
        os.mkdir("./{}/logos_bits".format(prefix))
        os.mkdir("./{}/logos_bits_no_axis".format(prefix))
        os.mkdir("./{}/logos_freq".format(prefix))
        os.mkdir("./{}/logos_freq_png".format(prefix))
        os.mkdir("./{}/logos_bits_png".format(prefix))
    else:
        os.system("rm -r ./{}/*".format(prefix))
        os.mkdir("./{}/fasta".format(prefix))
        os.mkdir("./{}/logos_bits".format(prefix))
        os.mkdir("./{}/logos_bits_no_axis".format(prefix))
        os.mkdir("./{}/logos_freq".format(prefix))
        os.mkdir("./{}/logos_freq_png".format(prefix))
        os.mkdir("./{}/logos_bits_png".format(prefix))
    if filter is not None:
        df = df[df[filter] == True].copy()
    clusters = set(df[cluster_col].tolist())
    for g in clusters:
        subdf = df[df[cluster_col] == g]
        with open("./{}/fasta/cluster_{}.fa".format(prefix, g), "w") as output:
            N = 0
            for idx, row in subdf.iterrows():
                output.write(">{}\n{}\n".format(idx, row[motif_column].replace("T", "U")))  # to RNA bases
                N += 1
            print("Cluster #{}: {}".format(g, N))
    if draw_logos == True:
        for g in clusters:
            os.system("weblogo -A rna -D fasta -F pdf --resolution 1000 --color-scheme classic --composition none -i -10 -P cluster_{g} -f ./{prefix}/fasta/cluster_{g}.fa > ./{prefix}/logos_bits/cluster_{g}.pdf".format(prefix=prefix, g=g))
            
            os.system("weblogo -A rna -D fasta -F png --resolution 1000 --color-scheme classic --composition none -i -10 -P cluster_{g} -f ./{prefix}/fasta/cluster_{g}.fa > ./{prefix}/logos_bits_png/cluster_{g}.png".format(prefix=prefix, g=g))

            os.system("weblogo -A rna -D fasta -F pdf -y Frequency --resolution 1000 --color-scheme classic --units probability --composition none -i -10 -P cluster_{g} -f ./{prefix}/fasta/cluster_{g}.fa > ./{prefix}/logos_freq/cluster_{g}.pdf".format(prefix=prefix, g=g))
            
            os.system("weblogo -A rna -D fasta -F png -y Frequency --resolution 1000 --color-scheme classic --units probability --composition none -i -10 -P cluster_{g} -f ./{prefix}/fasta/cluster_{g}.fa > ./{prefix}/logos_freq_png/cluster_{g}.png".format(prefix=prefix, g=g)) 
            
            os.system("weblogo -A rna -D fasta -X no -Y no -P \"\" -F pdf --resolution 1000 --color-scheme classic --composition none -i -10 -f ./{prefix}/fasta/cluster_{g}.fa > ./{prefix}/logos_bits_no_axis/cluster_{g}.pdf".format(prefix=prefix, g=g))

def hist_to_spots(hist2d, cutoff = 5, bins=[600, 600],
                 pixel_lower = 1, pixel_upper = 10,
                 show_small_clusters_id=True, show_big_clusters_id=True,
                 figsize=(12,12), figure_name="hist2D.png"):
    """This function is used for converting 2D histogram to spots (clusters).
   
    Parameters
    ----------
    hist2d: str
        The 2D histogram. (From draw_hist2d function)
    cutoff: int
        The cutoff for cv2.threshold, range from 0 to 255. Default=5.
    bins: tuple
        Should be equal to that of the hist2D.
    pixel_lower: int
        The lower limit of the pixels considering as a "small spot". Spots smaller than this will be ignored.
    pixel_upper: int
        The upper limit of the pixels considering as a "small spot". Spots larger than this will be considered as "big spot"
    show_small_clusters_id: boolean
        If draw the ids for small clusters.
    show_big_clusters_id: boolean
        If draw the ids for big clusters.
    figsize: tuple
        Figure size for matplotlib.
    figure_name: str
        The name of hist2D figure.
    Returns
    -------
    axes: matplotlib.axes
        The axes.
    dict_cnt_small:
        A dictionary of {id: locations} for the small spots.
    dict_cnt_big
        A dictionary of {id: locations} for the big spots.
    """

    import numpy as np
    import matplotlib.pyplot as plt
    import cv2
    from collections import defaultdict

    hist2d = hist2d.copy()
    cutoff = int(cutoff)

    hist2d = hist2d/(hist2d.max()/255.0)
    hist2d = hist2d.astype('uint8')
    hist2d_raw = hist2d.copy()
    
    thresh = cv2.threshold(hist2d, cutoff, 255, cv2.THRESH_BINARY)[1]
   
    thresh_small = hist2d_raw.copy()
    thresh_small[thresh_small>cutoff] = 0
    thresh_small[thresh_small>0] = 255
    
    contours, _ = cv2.findContours(thresh,cv2.RETR_LIST,cv2.CHAIN_APPROX_SIMPLE) # 
    contours_filtered_small = []
    contours_filtered_big = []
    dropped = []
    
    N = 0
    show_contours_small =np.zeros(bins).astype('uint8')
    show_contours_big =  np.zeros(bins).astype('uint8')
    dict_cnt_big = defaultdict(list)
    dict_cnt_small = defaultdict(list)
    
    mask =np.zeros(bins).astype('uint8')
    stistics = {"big_pixels":0, "big_cnt": 0, "small_pixels":0, "small_cnt":0, "ignored_pixels":0, "ignored_cnt":0}
    for cnt in contours:
        if pixel_lower <cv2.contourArea(cnt)< pixel_upper:
            contours_filtered_small.append(cnt)
            cimg = np.zeros(bins).astype('uint8')
            cv2.drawContours(cimg, [cnt], -1, 255, -1)
            cimg_rot90 = np.rot90(cimg, k = 1)
            pts = np.where(cimg_rot90 == 255)
            stistics["small_cnt"] += 1
            for i,j in zip(*pts):
                # c = hist2d_raw[i, j]
                stistics["small_pixels"] += 1
                dict_cnt_small[int(stistics["small_cnt"])].append((bins[0]-i, j))
        elif cv2.contourArea(cnt) >= pixel_upper:
            contours_filtered_big.append(cnt)
            cimg = np.zeros(bins).astype('uint8')
            cv2.drawContours(cimg, [cnt], -1, 255, -1)
            cimg_rot90 = np.rot90(cimg)
            pts = np.where(cimg_rot90 == 255)
            stistics["big_cnt"] += 1
            for i,j in zip(*pts):
                # c = hist2d_raw[i, j]
                stistics["big_pixels"] += 1
                dict_cnt_big[stistics["big_cnt"]].append((bins[0]-i, j))
        else:
            dropped.append(cnt)
            for item in cnt:
                stistics["ignored_cnt"] += 1
                for x,y in item:
                    c = hist2d_raw[x, y]
                    stistics["ignored_pixels"] += c
                        
    print("Pixels ignored: {} pixels in {} cntours".format(stistics["ignored_pixels"], stistics["ignored_cnt"]))
    print("Big contours: {} pixels in {} cntours".format(stistics["big_pixels"], stistics["big_cnt"]))
    print("Small contour: {} piexels in {} cntours".format(stistics["small_pixels"], stistics["small_cnt"]))
    image = cv2.bitwise_and(thresh, thresh, mask=mask)

    cv2.drawContours(show_contours_small, contours_filtered_small, -1, 255, -1)
    cv2.drawContours(show_contours_big, contours_filtered_big, -1, 255, -1)
    cv2.drawContours(thresh_small, dropped, -1, 255, -1)

    fig, axes = plt.subplots(2,2, sharex=True, sharey=True,figsize=figsize)
    axes[0][0].imshow(thresh_small)
    axes[0][1].imshow(show_contours_small)
    axes[1][0].imshow(show_contours_big)
    axes[1][1].imshow(thresh)
    axes[0][0].set_title("Ignored pixels")
    axes[0][1].set_title("Small clusters")
    axes[1][0].set_title("Big clusters")
    axes[1][1].set_title("Thresh")
    axes[0][0].set_xticks([]) # 
    axes[0][0].set_yticks([]) #
    
    if show_big_clusters_id == True:
        for id, xy in dict_cnt_big.items():
            Xs = []
            Ys = []
            for x, y in xy:
                Xs.append(x)
                Ys.append(y)
            c_X = np.mean(Xs)
            c_Y = np.mean(Ys)
            
            diameter_X = np.max(Xs) - np.min(Xs)
            diameter_Y = np.max(Ys) - np.min(Ys)
            
            if diameter_X * diameter_Y >= 250:
                color = "red"
            else:
                color = "white"
            axes[1][0].annotate(str(id), xy=(c_X, c_Y), c=color, ha="center", va="center") 
    
    if show_small_clusters_id == True: 
        for id, xy in dict_cnt_small.items():
            Xs = []
            Ys = []
            for x, y in xy:
                Xs.append(x)
                Ys.append(y)
            c_X = np.mean(Xs)
            c_Y = np.mean(Ys)
            
            diameter_X = np.max(Xs) - np.min(Xs)
            diameter_Y = np.max(Ys) - np.min(Ys)
            
            if diameter_X * diameter_Y >= 250:
                color = "red"
            else:
                color = "white"
            axes[0][1].annotate(str(id), xy=(c_X, c_Y), c=color, ha="center", va="center") 
            
    plt.tight_layout()
    plt.savefig(figure_name, dpi=300)
    return axes, dict_cnt_small, dict_cnt_big

def retrive_clusters(df, edgesX, edgesY, dict_clusters, bins=[600,600], cluster_ids=None, spot_name="spot"):
    """This function is used for annotate sites with clusters.
   
    Parameters
    ----------
    df: pd.DataFrame
        A DataFrame object with X and Y column.
    edgesX: np.array
        The array of X edges generated by hist2D.
    edgesY: np.array
        The array of Y edges generated by hist2D.
    dict_clusters: dict
        The dictionary generated by hist_to_spots.
    bins: tuple
        Should equal to that of hist2D.
    cluster_ids: iterable
        If given, only find clusters with that ids.
    spot_name: str
        The column name of the clusters.
    Returns
    -------
    pd.DataFrame
    """
    import sys

    df = df.copy()
    indexes = []
    spots  = []
    N = 0
    M = 0
    S = 0
    for spot_id, xy in dict_clusters.items():
        if cluster_ids is None or spot_id in cluster_ids:
            N += 1
            for x, y in xy:
                left = edgesX[x]
                right = edgesX[x+1]
                bottom = edgesY[bins[1]-y-1]
                top = edgesY[bins[1]-y]
                slice = df.loc[(df["X"]>left)& (df["X"]<=right) & (df["Y"]>bottom) & (df["Y"]<=top)]
                indexes.extend(slice.index.tolist())
                spots.extend([spot_id] * slice.shape[0])
                M += 1
                S += slice.shape[0]
                sys.stdout.write("Number of spots: {}, Total pixels:{}, Total sites: {}\r".format(N, M, S))
    df.loc[indexes, "Spot_name"] = spot_name
    df.loc[indexes, "Spot_id"] = spots
    return df

def prepare_kmers_dict(df, column="motif_F14"):
    """Prepare all kmers from a DataFrame with flanking 14 nt sequences.
   
    Parameters
    ----------
    df: pd.DataFrame
        A DataFrame object with X and Y column.
    column: str
        The name of column containing flanking 14 nt sequences.
    Returns
    -------
    dict
    """
    import pandas as pd
    dict_all_5mers = {-4: {}, -3: {}, -2: {}, -1: {}, 0: {}, 1: {}, 2: {}, 3: {}, 4: {}}
    indexes = []
    
    for idx, row in df.iterrows():
        seq = str(row[column]).upper()
        if  len(seq) != 29:
            continue
        for i in range(0, 9):
            kmers_id = i - 4
            dict_all_5mers[kmers_id][idx] = seq[i: i+21]
    return dict_all_5mers

def phase_shift(df, dict_all_5mers, cluster_id=None, column_motif_F10="motif_F10", current_phase=0, target_base="A"):
    """Perform phase shift.
   
    Parameters
    ----------
    df: pd.DataFrame
        A DataFrame object with X and Y column.
    dict_all_5mers: dict
        The dictionary from prepare_kmers_dict.
    cluster_id: int
        The id of specific cluster.
    column_motif_F10: str
        The column name of the 10-nt flanking sequences.
    current_phase: int
        The current phase of the cluster.
    target_base: str
        The target base to perform phase matching.
    Returns
    -------
    pd.DataFrame
    """
    df = df.copy()
    indexes = []
    shifted_seq = []
    for idx, row in df.iterrows():
        if row["Cluster"]==cluster_id and row["Center_base"]!=target_base:
            if row[column_motif_F10][10] != target_base and row[column_motif_F10][10+current_phase] == target_base:
                indexes.append(idx)
                shifted_seq.append(dict_all_5mers[current_phase][idx])
    df.loc[indexes, "Phase_shift"] = current_phase
    df.loc[indexes, column_motif_F10] = shifted_seq
    df.loc[indexes, "Center_base"] = target_base
    return df

if __name__ == "__main__":
    pass


def run_HDBSCAN(df=None, X=None, Y=None, soft_clustering=True, min_cluster_size=100, min_samples=10,
                cluster_selection_epsilon=0.0, cluster_selection_method='eom',
                draw_condensed_tree=True, core_dist_n_jobs=6):
    """An implement of HDBSCAN (CPU version)
   
    Parameters
    ----------
    df: pd.DataFrame
        A DataFrame with columns X and Y.
    X: iterable
        A list of X values.
    Y: iterable
        A list of Y values.
    soft_clustering: boolean
        Use soft clustering or not. Default=True.
    min_cluster_size: int
        min_cluster_size in HDBSCAN.
    min_samples: int
        min_samples in HDBSCAN
    cluster_selection_epsilon: float
        cluster_selection_epsilon in HDBSCAN
    cluster_selection_method: str
        cluster_selection_method in HDBSCAN. Should be "eom" or "leaf".
    draw_condensed_tree: boolean
        Draw the condensed tree of HDBSCAN or not.
    core_dist_n_jobs:
        core_dist_n_jobs in HDBSCAN.
    Returns
    -------
    sequences_onehot: list
        A list of one-hot encoded sequences.
    """
    import numpy as np
    import hdbscan
    from collections import Counter
    import matplotlib.pyplot as plt
    import seaborn as sns

    if df is None and X is None and Y is None:
        raise ValueError("Please provide a DataFrame or a paired X and Y!")
    if df is not None:
        df = df.copy()
        INPUT = np.stack([df["X"], df["Y"]], axis=1)
    
    if soft_clustering == True:
        model = hdbscan.HDBSCAN(min_cluster_size=min_cluster_size, min_samples=min_samples, cluster_selection_epsilon=cluster_selection_epsilon, cluster_selection_method=cluster_selection_method, core_dist_n_jobs=core_dist_n_jobs, prediction_data=True)
        yhat = model.fit(INPUT)
        soft_clusters = hdbscan.all_points_membership_vectors(yhat)
        labels = [np.argmax(x) for x in soft_clusters] 
    else:
        model = hdbscan.HDBSCAN(min_cluster_size=min_cluster_size, min_samples=min_samples, cluster_selection_epsilon=cluster_selection_epsilon, cluster_selection_method=cluster_selection_method, core_dist_n_jobs=core_dist_n_jobs, prediction_data=False)
        labels = model.fit_predict(INPUT)
    
    clusters = [i+1 if i > -1 else -1 for i in labels ]  # re-number lables to make it human-readable
    print("HDBSCAN cluster number: {}".format(np.max(clusters)))

    if draw_condensed_tree == True:
        fig, ax = plt.subplots()
        model.condensed_tree_.plot(select_clusters=True, selection_palette=sns.color_palette())
        plt.savefig("Condensed_tree.pdf")

    if df is not None:
        df["Cluster"] = clusters
        print(df.groupby("Cluster")["Cluster"].count())
        return df, model
    else:
        for i, j in Counter(clusters).items():
            print(i, j)
        return clusters, model

def run_HDBSCAN_GPU(df=None, X=None, Y=None, min_cluster_size=100, min_samples=10,
                cluster_selection_epsilon=0.0, cluster_selection_method='eom'
                ):
    """An implement of HDBSCAN (GPU version). Only standard clustering mode is available.
   
    Parameters
    ----------
    df: pd.DataFrame
        A DataFrame with columns X and Y.
    X: iterable
        A list of X values.
    Y: iterable
        A list of Y values.
    min_cluster_size: int
        min_cluster_size in HDBSCAN.
    min_samples: int
        min_samples in HDBSCAN
    cluster_selection_epsilon: float
        cluster_selection_epsilon in HDBSCAN
    cluster_selection_method: str
        cluster_selection_method in HDBSCAN. Should be "eom" or "leaf".

    Returns
    -------
    sequences_onehot: list
        A list of one-hot encoded sequences.
    """
    import numpy as np
    import cuml
    from collections import Counter
    import matplotlib.pyplot as plt

    if df is None and X is None and Y is None:
        raise ValueError("Please provide a DataFrame or a paired X and Y!")
    if df is not None:
        df = df.copy()
        INPUT = np.stack([df["X"], df["Y"]], axis=1)

    model = cuml.HDBSCAN(min_cluster_size=min_cluster_size, min_samples=min_samples, cluster_selection_epsilon=cluster_selection_epsilon, cluster_selection_method=cluster_selection_method, prediction_data=False)
    labels = model.fit_predict(INPUT)
    
    clusters = [i+1 if i > -1 else -1 for i in labels ]  # re-number lables to make it human-readable
    print("HDBSCAN cluster number: {}".format(clusters.max()))

    if df is not None:
        df["Cluster"] = clusters
        print(df.groupby("Cluster")["Cluster"].count())
        return df, model
    else:
        for i, j in Counter(clusters).items():
            print(i, j)
        return clusters, model

def run_HDBSCAN_subclustering(df=None, target=None, cluster_col="Cluster", soft_clustering=True,
                min_cluster_size=100, min_samples=10,
                cluster_selection_epsilon=0.0, cluster_selection_method='eom',
                draw_condensed_tree=True, core_dist_n_jobs=None):
    """An implement of HDBSCAN (CPU version) for further clustering of a subcluster.
   
    Parameters
    ----------
    df: pd.DataFrame
        A DataFrame with columns X, Y, and clusters.
    soft_clustering: boolean
        Use soft clustering or not. Default=True.
    min_cluster_size: int
        min_cluster_size in HDBSCAN.
    min_samples: int
        min_samples in HDBSCAN
    cluster_selection_epsilon: float
        cluster_selection_epsilon in HDBSCAN
    cluster_selection_method: str
        cluster_selection_method in HDBSCAN. Should be "eom" or "leaf".
    draw_condensed_tree: boolean
        Draw the condensed tree of HDBSCAN or not.
    core_dist_n_jobs:
        core_dist_n_jobs in HDBSCAN.
    Returns
    -------
    sequences_onehot: list
        A list of one-hot encoded sequences.
    """
    import numpy as np
    import hdbscan
    from collections import Counter
    import matplotlib.pyplot as plt
    import seaborn as sns

    df = df.copy()
    max_cluster_id = df[cluster_col].max()
    
    df1 = df[df[cluster_col]==target].copy()
    
    X = np.stack([df1["X"], df1["Y"]], axis=1)
    
    model = hdbscan.HDBSCAN(min_cluster_size=min_cluster_size, min_samples=min_samples, cluster_selection_method=cluster_selection_method, cluster_selection_epsilon=cluster_selection_epsilon, core_dist_n_jobs=core_dist_n_jobs, prediction_data=True)
    yhat = model.fit(X)
    
    soft_clusters = hdbscan.all_points_membership_vectors(yhat)
    labels = [np.argmax(x) for x in soft_clusters] 

    df1[cluster_col] = [max_cluster_id + i + 1 for i in labels ]  # re-number lables to make it human-readable
    df.loc[df1.index, cluster_col] = df1[cluster_col].tolist()

    print("HDBSCAN cluster number: {}".format(df["Cluster"].max()-1))
    print(df.groupby(cluster_col)[cluster_col].count())

    if draw_condensed_tree == True:
        fig, ax = plt.subplots()
        model.condensed_tree_.plot(select_clusters=True, selection_palette=sns.color_palette())
        plt.savefig("Condensed_tree_subcluster.pdf")
    return df, model

def run_Louvain(graph, df=None, random_state=42, resolution_parameter=1.0):
    """Clustering UMAP result with Louvain.
   
    Parameters
    ----------
    graph: iGraph object
        An iGraph object computed from UMAP nearest neighbor.
    df: pd.DataFrame
        If given, will add a column named "Cluster" to the DataFrame; else will return the labels.
    random_state: int
        Random seed.
    resolution_parameter: float
        resolution_parameter for Louvain
    
    Returns
    -------
    pd.DataFrame or a list
    """

    import louvain
    import numpy as np

    partition_type = louvain.RBConfigurationVertexPartition
    model = louvain.find_partition(graph, partition_type, seed=random_state, resolution_parameter=resolution_parameter)
    labels = np.array(model.membership)
    
    if df is not None:
        df = df.copy()
        df["Cluster"] = labels
        return df
    else:
        return labels

def run_Leiden(graph, df=None, random_state=42, resolution_parameter=1.0):
    """Clustering UMAP result with Leiden.
   
    Parameters
    ----------
    graph: iGraph object
        An iGraph object computed from UMAP nearest neighbor.
    df: pd.DataFrame
        If given, will add a column named "Cluster" to the DataFrame; else will return the labels.
    random_state: int
        Random seed.
    resolution_parameter: float
        resolution_parameter for Louvain
    
    Returns
    -------
    pd.DataFrame or a list
    """

    import leidenalg
    import numpy as np

    partition_type = leidenalg.RBConfigurationVertexPartition
    model = leidenalg.find_partition(graph, partition_type, seed=random_state, resolution_parameter=resolution_parameter)
    labels = np.array(model.membership)
    
    if df is not None:
        df = df.copy()
        df["Cluster"] = labels
        return df
    else:
        return labels

if __name__ == "__main__":
    pass


def onehot_encoder_df(df, column="seq", enc_bases="ATCGN"):
    """This function is used for generate One-Hot encoding sequences from a DataFrame.
   
    Parameters
    ----------
    df: pd.DataFrame
        A DataFrame.
    column: str or tuple
        The column containing the sequences
    enc_bases: str
        The encoding bases. Default="ATCGN".

    Returns
    -------
    sequences_onehot: list
        A list of one-hot encoded sequences.
    """
    from sklearn.preprocessing import OneHotEncoder
    import numpy as np
    enc = OneHotEncoder(dtype=np.int8)
    enc.fit([[i] for i in enc_bases])
    
    sequences_onehot = []
    for idx, row in df.iterrows():
        seq = [[i] for i in str(row[column]).upper()]
        sequences_onehot.append(enc.transform(seq).toarray().reshape(-1))
    return sequences_onehot

def onehot_encoder_iterable(iter_obj, enc_bases="ATCGN"):
    """This function is used for generate One-Hot encoding sequences from a iterable object.
   
    Parameters
    ----------
    iter_obj: iterable
        An iterable object containing the sequences.
    enc_bases: str
        The encoding bases. Default="ATCGN".
    Returns
    -------
    sequences_onehot: list
        A list of one-hot encoded sequences.
    """
    from sklearn.preprocessing import OneHotEncoder
    import numpy as np
    enc = OneHotEncoder(dtype=np.int8)
    enc.fit([[i] for i in enc_bases])
    
    sequences_onehot = []
    for item in iter_obj:
        seq = [[i] for i in item.upper()]
        sequences_onehot.append(enc.transform(seq).toarray().reshape(-1))
    return sequences_onehot

def run_UMAP(onehot_input, df=None, init="random", random_state=42, min_dist=0.01, n_neighbors=20, densmap=False, verbose=True, n_jobs=6):
    """An implement of UMAP (CPU version).
   
    Parameters
    ----------
    onehot_input: iterable.
        A list of one-hot encoded sequences.
    df: pd.DataFrame
        A DataFrame to process. If given, it will return a DataFrame with X and Y columns. If not, it will return X and Y, separatively.
    init: str.
        init value for UMAP.
    random_state: int
        random seed.
    min_dist: float
        min_dist for UMAP
    n_neighbors: int
        n_neighbors for UMAP
    densmap: boolean
        If use DensMAP.
    verbose: boolean
        verbose level

    Returns
    -------
    A DataFrame or [X and Y]
    """
    import umap
   
    model = umap.UMAP(init=init, random_state=random_state, n_components=2, min_dist=min_dist, n_neighbors=n_neighbors, verbose=verbose, densmap=densmap, n_jobs=n_jobs)
    umap_output = model.fit_transform(onehot_input)
    
    if df is not None:
        df = df.copy()
        df["X"] = umap_output[:, 0]
        df["Y"] = umap_output[:, 1]
        del model
        return df
    else:
        del model
        return umap_output[:, 0], umap_output[:, 1]
    
def run_UMAP_GPU(onehot_input, df=None, init="random", random_state=42, min_dist=0.01, n_neighbors=20, densmap=False, verbose=True):
    """An implement of UMAP (GPU version).
   
    Parameters
    ----------
    onehot_input: iterable.
        A list of one-hot encoded sequences.
    df: pd.DataFrame
        A DataFrame to process. If given, it will return a DataFrame with X and Y columns. If not, it will return X and Y, separatively.
    init: str.
        init value for UMAP.
    random_state: int
        random seed.
    min_dist: float
        min_dist for UMAP
    n_neighbors: int
        n_neighbors for UMAP
    densmap: boolean
        If use DensMAP.
    verbose: boolean
        verbose level

    Returns
    -------
    A DataFrame or [X and Y]
    """
    import cuml
   
    model = cuml.UMAP(init=init, random_state=random_state, n_components=2, min_dist=min_dist, n_neighbors=n_neighbors, verbose=verbose, densmap=densmap)
    umap_output = model.fit_transform(onehot_input)
    
    if df is not None:
        df = df.copy()
        df["X"] = umap_output[:, 0]
        df["Y"] = umap_output[:, 1]
        del model
        return df
    else:
        del model
        return umap_output[:, 0], umap_output[:, 1]

def get_sparse_matrix_from_indices_distances_umap(
        knn_indices, knn_dists, n_obs, n_neighbors
    ):
    """A helper function for Louvain and Leiden. Adopted from Scanpy.
   
    Parameters
    ----------
    knn_indices: object
    knn_dists: object
    n_obs: int
    n_neighbors: int
    """
    import numpy as np
    from scipy.sparse import issparse, coo_matrix, csr_matrix
    rows = np.zeros((n_obs * n_neighbors), dtype=np.int64)
    cols = np.zeros((n_obs * n_neighbors), dtype=np.int64)
    vals = np.zeros((n_obs * n_neighbors), dtype=np.float64)

    for i in range(knn_indices.shape[0]):
        for j in range(n_neighbors):
            if knn_indices[i, j] == -1:
                continue  # We didn't get the full knn for i
            if knn_indices[i, j] == i:
                val = 0.0
            else:
                val = knn_dists[i, j]

            rows[i * n_neighbors + j] = i
            cols[i * n_neighbors + j] = knn_indices[i, j]
            vals[i * n_neighbors + j] = val

    result = coo_matrix((vals, (rows, cols)), shape=(n_obs, n_obs))
    result.eliminate_zeros()
    return result.tocsr()

def compute_connectivities_umap(
        knn_indices,
        knn_dists,
        n_obs,
        n_neighbors,
        set_op_mix_ratio=1.0,
        local_connectivity=1.0,
    ):
    """A helper function for Louvain and Leiden. Adopted from Scanpy.
   
    Parameters
    ----------
    knn_indices: object
    knn_dists: object
    n_obs: int
    n_neighbors: int
    set_op_mix_ratio: float
    local_connectivity: float
    """
    from scipy.sparse import issparse, coo_matrix, csr_matrix
    from umap.umap_ import fuzzy_simplicial_set
    X = coo_matrix(([], ([], [])), shape=(n_obs, 1))
    connectivities = fuzzy_simplicial_set(
        X,
        n_neighbors,
        None,
        None,
        knn_indices=knn_indices,
        knn_dists=knn_dists,
        set_op_mix_ratio=set_op_mix_ratio,
        local_connectivity=local_connectivity,
    )

    if isinstance(connectivities, tuple):
        # In umap-learn 0.4, this returns (result, sigmas, rhos)
        connectivities = connectivities[0]

    distances = get_sparse_matrix_from_indices_distances_umap(
        knn_indices, knn_dists, n_obs, n_neighbors
    )

    return distances, connectivities.tocsr()

def get_igraph_from_adjacency(adjacency, directed=None):
    """A helper function for Louvain and Leiden. Adopted from Scanpy.
   
    Parameters
    ----------
    adjacency: object
        Generated by compute_connectivities_umap
    
    Returns
    -------
    iGraph object
    """
    import numpy as np
    import igraph as ig
    sources, targets = adjacency.nonzero()
    weights = adjacency[sources, targets]
    if isinstance(weights, np.matrix):
        weights = weights.A1
    g = ig.Graph(directed=directed)
    g.add_vertices(adjacency.shape[0])  # this adds adjacency.shape[0] vertices
    g.add_edges(list(zip(sources, targets)))
    try:
        g.es['weight'] = weights
    except KeyError:
        pass
    return g

def get_igraph(onehot_input, random_state=42, metric="euclidean", n_neighbors=20, metric_kwds={}, n_jobs=6, angular=False, verbose=False):
    """Prepare iGraph object for Louvain and Leiden
   
    Parameters
    ----------
    onehot_input: np.array
        The one-hot encoded sequences.
    random_state: int
        Random seed.
    metric: str
        Same as UMAP performed.
    n_neighbors: int
        Same as UMAP.
    metric_kwds: dict
    angular: boolean
    verbose: boolean
    
    Returns
    -------
    iGraph object
    """

    from umap.umap_ import nearest_neighbors

    n_obs = onehot_input.shape[0]
    
    knn_indices, knn_dists, forest = nearest_neighbors(
            onehot_input,
            n_neighbors,
            random_state=random_state,
            metric=metric,
            metric_kwds=metric_kwds,
            angular=angular,
            verbose=verbose,
            n_jobs=n_jobs,
        )
      
    distances, connectivities = compute_connectivities_umap(knn_indices, knn_dists, n_obs, n_neighbors, set_op_mix_ratio=1.0, local_connectivity=1.0)
 
    g = get_igraph_from_adjacency(connectivities)
    return g
    
if __name__ == "__main__":
    pass


import time
import sys, os
import numpy as np
import pandas as pd
from sklearn.preprocessing import OneHotEncoder
import scipy.stats
import tracemalloc
import umap
import hdbscan
from inspect import Parameter
import weblogo
from weblogo import *

def onehot_enc(row, expected_length=21):
    """Encode the data with one-hot encoding
    Parameter:
    ---------
        row: string
            DNA or RNA sequence for one-hot encoding. 
    ---------

    Retruns:
    --------
        onehot_array: array
            one-hot encoding data 
    """
    enc = OneHotEncoder(dtype=np.int8)
    enc.fit([[i] for i in "ATCGN"])
    seq = [[i] for i in row["seq"].upper().replace("U","T") if len(row["seq"]) == expected_length]
    onehot_array = enc.transform(seq).toarray().reshape(-1)
    return onehot_array

def UMAP(onehot_input, df, parameters):
    """Dimensionality reduction with UMAP

    Parameter:
    ---------
        onehot_input: array
            One-hot encoding data from sequence.
        df: pd.DataFrame
            Sequence in pd.DataFrame format. It will be added two columns named "X" and "Y" as results of UMAP. 
    ---------
    Retruns:
        pd.DataFrame
    --------
    """
    df = df.copy()

    random_state = int(parameters["random_state"])
    init = str(parameters["umap_init"])
    min_dist = float(parameters["min_dist"])
    n_neighbors = int(parameters["n_neighbors"])
    densmap = bool(parameters["densmap"])
    n_jobs = int(parameters["umap_jobs"])

    model = umap.UMAP(init=init, random_state=random_state, n_components=2, min_dist=min_dist, n_neighbors=n_neighbors, verbose=True, densmap=densmap, n_jobs=n_jobs)
    umap_output = model.fit_transform(onehot_input)
    
    df["X"] = umap_output[:, 0]
    df["Y"] = umap_output[:, 1]
    
    del model
    return df

def cluster_HDBSCAN(df, parameters):
    """Clustering UMAP results with HDSCAN

    Parameter:
    ---------
        df: pd.DataFrame
            The results of UMAP computed from UMAP nearst neighbor. After clustering with HDBSCAN, add a columns to it named "Cluster".
        parameters: list
            Parameters for HDBSCAN from user. 
    ---------

    Retruns:
    --------
        pd.DataFrame 
    """
    # use multi-code here
    df = df.copy()
    X = np.stack([df["X"], df["Y"]], axis=1)

    min_cluster_size = int(parameters["min_cluster_size"])
    min_samples = int(parameters["min_samples"])
    cluster_selection_method = str(parameters["cluster_selection_method"])
    core_dist_n_jobs = int(parameters["hdbscan_jobs"])
    cluster_selection_epsilon = float(parameters["cluster_selection_epsilon"])
    if bool(parameters["softclustering"]) == True:
        prediction_data = True
    else:
        prediction_data = False

    model = hdbscan.HDBSCAN(min_cluster_size = min_cluster_size, min_samples=min_samples, cluster_selection_method=cluster_selection_method, cluster_selection_epsilon=cluster_selection_epsilon, core_dist_n_jobs=core_dist_n_jobs, prediction_data=prediction_data)
    if prediction_data == True:
        yhat = model.fit(X)
        soft_clusters = hdbscan.all_points_membership_vectors(yhat)
        labels = [np.argmax(x) for x in soft_clusters] 
    else:
        labels = model.fit_predict(X)
    
    df["Cluster"] = [i+1 if i > -1 else -1 for i in labels ]  # re-number lables to make it human-readable

    # check cluster number
    # print(df.groupby("Cluster")["Cluster"].count())
    return df


def run_cluster(fasta_df, path, parameters):
    """Call two algorithms to cluster and get the result  
    Parameter:
    ---------
        fasta_df: pd.DataFrame
            Sequence file in pd.DataFrame format. After calling two algorithms, add three columns to it named "X", "Y" and "Cluster" to store the results.
        path: str
            The output directory.
        parameters: list
            Parameters for HDBSCAN from user input.

    ---------
    Retruns:
    --------
        pd.DataFrame
    """
    onehot_input = []
    for idx, row in fasta_df.iterrows():
        onehot_input.append(onehot_enc(row, expected_length=int(parameters["exp_len"])))
    onehot_input = np.array(onehot_input)

    df_UMAP = UMAP(onehot_input, fasta_df, parameters)
    df_HDBSCAN = cluster_HDBSCAN(df_UMAP, parameters)
    # print(df_HDBSCAN)
    df_HDBSCAN.to_csv("{path}/all_clusters.csv".format(path=path),index = None)
    
    base_type = parameters["weblogo_base_type"]

    with open("{path}/init.fa".format(path=path), "w") as init_fasta:
        for idx, row in df_HDBSCAN.iterrows():
            if base_type == "DNA":
                seq_out = str(row["seq"]).upper().replace("U", "T")
            elif base_type == "RNA":
                seq_out = str(row["seq"]).upper().replace("T", "U")
            else:
                seq_out = str(row["seq"]).upper()
            init_fasta.write(">{}\n{}\n".format(idx, seq_out))
    return df_HDBSCAN

def draw_logo(infile, parameters):
    """Create sequence logos with Weblogo
    Parameters:
    ---------
        infile: string
            A sequence file in fasta format to create sequence logo with Weblogo.
    ---------
    Returns:
        A picture of sequence logo in png format.
    --------
    """
    
    unit = parameters["weblogo_unit"]
    first_index = int(parameters["weblogo_first_index"])
    
    data = open(infile)
    seqs = read_seq_data(data)

    logodata = LogoData.from_seqs(seqs)
    logooptions = LogoOptions(
        unit_name = unit, # 'probability',
        yaxis_label = unit, # 'probability',
        first_index = first_index, # -10,
        color_scheme = classic,
        resolution = 1000
    )
    logoformat = LogoFormat(logodata,logooptions)
    png = png_print_formatter(logodata,logoformat)

    return png




#! /usr/bin/env python3
from ensurepip import version
import sys, os
import argparse
from iMVP_utils.interactive import launch_backend

if __name__ == "__main__":
    """The iMVP modification viewer application.
    Parameters
    ----------
    --output, -out: str
        The path used for storing the files (figures, fasta, etc.).
    --host: str
        The IP address of the sever.
    --port: int
        The port id.
    """
    parser = argparse.ArgumentParser(prog="iMVP_viewer")
    parser.add_argument('--output','-o', dest="output_path", type = str, default="./output/",help='The output directory. If not exist, will make the folder. Default="./output/"')
    parser.add_argument('--host', dest="host", type = str, default="127.0.0.1",help='The IP address of the app, default=127.0.0.1 (localhost)')
    parser.add_argument('--port', dest="port", type = int, default=8050,help='The port number that you want to dispaly the app.')
    parser.add_argument('--version', '-v', action='version', version="0.2.3",help='Display version.')

    options = parser.parse_args()

    host_ip_address = options.host
    port_id = options.port
    output_path = options.output_path
    app = launch_backend(output_path=output_path)

    app.run_server(debug=True, host=host_ip_address, port=port_id) 


import os
import io
from pydoc import classname
from Bio import SeqIO
import time
import iMVP_utils
from iMVP_utils import interactive_functions
# import interactive_functions
import base64
import PIL.Image as Image
import pandas as pd
import numpy as np
import dash
from dash import dcc
from dash import html
from dash.dependencies import Input, Output,State
from dash import callback_context
from flask import Flask
import plotly.express as px
import plotly.graph_objects as go

def launch_backend(output_path="./output/"):
    """The function to launch the backend for interactive

    Parameters
    ---------
    output_path: str
        The output directory of the files.
    
    Returns
    ---------
        A Dash App object.
    """
    iMVP_finished = False
    input_validated = False
    first_time = True

    assets_path = os.path.dirname(iMVP_utils.__file__) + "/assets/"
    
    server= Flask(__name__)
    app = dash.Dash(name="app1", server=server, assets_folder=assets_path)

    """Run this app with 'python app.py -p port_number and visit http://127.0.0.1:prot_number/ in your web browser.
    (Press CTRL+C to quit)

    """

    if os.path.exists (output_path) == False:
        os.mkdir(output_path)

    def run_iMVP(content, input_parameters):
        """Clustering upload data in fasta format with UMAP and HDBSCAN. 

        Parameters
        ---------
        content: string
            A comma separated string, including type and content of upload file. 
        input_parameters: dict
            A list of reserved parameters for HDBSCAN.
        ---------

        Returns
        ---------
            A Div component: chilren
                A html div of 'processing' information.
            Html stlye: dict
                A style of 'submit-data' button. 
            HDBSCAN_dict: dict
                The results of HDBSCAN.
        """
        nonlocal output_path
        time_start = time.time()
        _, content_string = content.split(',')
        
        decoded = base64.b64decode(content_string)
        style_submit_button = {'width': '40%'}
    
        df = pd.read_csv(
            io.StringIO(decoded.decode('utf-8').replace("\r\n", "\n").replace("\r", "\n").replace(">", "")),sep = "\n", header=None)
        fa_data = pd.concat([df[::2].reset_index(drop=True),df[1::2].reset_index(drop=True)],axis=1)
        fa_data.columns = ['sites','seq']
        # run HDBSACN
        df_HDBSCAN = interactive_functions.run_cluster(fa_data, output_path, input_parameters)
        
        png = interactive_functions.draw_logo("{path}/init.fa".format(path=output_path), input_parameters)

        image = Image.open(io.BytesIO(png))
        img_file = '{path}/weblogo.png'.format(path=output_path)
        image.save(img_file) 

        HDBSCAN_dict = df_HDBSCAN.to_dict()
        
        time_end = time.time()
        
        
        # used_time = round((time_end - time_start)/60,2)
        used_time = time_end - time_start
        if used_time >= 3600:
            used_time = time.strftime("%H hr %M min %S sec", time.gmtime(used_time))    
        elif used_time >= 60:
            used_time = time.strftime("%M min %S sec", time.gmtime(used_time))   
        else:
            used_time = time.strftime("%S sec", time.gmtime(used_time))   
            
        # except Exception as e :
        #     return html.Div([
        #         'There was an error processing this file.'
        #     ]), {'display':'none'}

        df_groupby = df_HDBSCAN.groupby("Cluster")[["Cluster"]].count()
        df_groupby.columns = ["Count"]
        df_groupby["Cluster ID"] = df_groupby.index
        df_groupby = df_groupby[["Cluster ID", "Count"]]
        print(df_groupby)

        # html.Div([
        #     dash.dash_table.DataTable(df_groupby.to_dict('records'), [{"name": i, "id": i} for i in df_groupby.columns])
        # ], style={"width": "400px", "margin-left":"50px"}
        # ),

        return html.Div([html.H3("{} inputs were analyzed. Finished in {}!".format(df_HDBSCAN.shape[0], used_time)), html.H3('Done!'),]), \
        style_submit_button , HDBSCAN_dict

    def check_fasta_input(content):
        basespace = {i:1 for i in "ATCGUNRDEQHILKMFPSWYV"}
        _, content_string = content.split(',')
        decoded = base64.b64decode(content_string)
        N = 0
        is_fasta = True
        try:
            for line in decoded.decode('utf-8').replace("\r\n", "\n").replace("\r", "\n").split("\n"):
                try:
                    # print(line)
                    if not line:
                        continue
                    elif N % 2 == 0:
                        if line.startswith(">") == False:
                            is_fasta = False
                            break
                    elif N % 2 == 1:
                        for b in line:
                            if b.upper() not in basespace:
                                is_fasta = False
                                break
                    N += 1
                except Exception as e:
                    is_fasta = False
                    break
        except Exception as e:
            is_fasta = False
        return is_fasta

    @app.callback(
        Output('upload_data','children'),
        Input('upload_data','filename'),
        State('upload_data', 'contents')
    )

    def upload_info(filenames, contents):
        nonlocal input_validated
        input_validated = False
        if filenames is None:
            return html.Div([
                'Drag and Drop or ',html.A('Click to upload your FASTA file here')
            ], style={"line-height": "300px", "width": "600px"}) #  style={"padding":"50px"}
        else:
            for file, c in zip(filenames, contents):
                fasta_status = check_fasta_input(c)
                if fasta_status == True:
                    input_validated = True
                    return html.Div([html.B('{filename}'.format(filename = file),style={'color':'#ff4777'}),' has been uploaded'], style={"line-height": "300px", "width": "600px"})
                else:
                    return html.Div([html.B('Warining: {filename} is not a validated FASTA file!'.format(filename = file),style={'color':'#ff4777'}),], style={"line-height": "300px", "width": "600px"})
    @app.callback(
                Output('hdbscan-parameter-list','data'),
                Output('processing_1', 'children'),
                Output('processing_3','children'),
                Output('upload_data', 'style'),
                Output('para-list-div', 'style'),
                Output('upload-div', 'style'),
                Input('submit-button-state', 'n_clicks'),
                [
                State('upload_data', 'contents'), 
                State('exp_len', 'value'), 
                State('n_neighbors', 'value'), 
                State('min_dist', 'value'), 
                State('random_state', 'value'), 
                State('umap_jobs', 'value'), 
                State('umap_init', 'value'), 
                State('densmap', 'value'), 
                State('min_cluster_size', 'value'),
                State('min_samples', 'value'),
                State('cluster_selection_method', 'value'),
                State('cluster_selection_epsilon', 'value'),
                State('hdbscan_jobs', 'value'),
                State('softclustering', 'value'),
                State('weblogo_unit', 'value'),
                State('weblogo_first_index', 'value'),
                State('weblogo_base_type', 'value'),
                ],
                prevent_initial_call=True) 

    def infomation_hide_div(n_clicks,list_of_contents, exp_len, n_neighbors, min_dist, random_state, umap_jobs, umap_init, densmap ,min_cluster_size, min_samples, cluster_selection_method, cluster_selection_epsilon, hdbscan_jobs, softclustering, weblogo_unit, weblogo_first_index, weblogo_base_type):
        """
        Parameters
        ---------
            n_clicks: int
                The number of clicks to trigger the clustering.
            list_of_contents: list
                Contents of upload data.
            exp_len: int
                The expected lengths for the sequences.
            n_neighbors: int
                n_neighbors for UMAP.
            min_dist: int
                min_dist for UMAP.
            random_state: int
                random_state for UMAP.
            umap_jobs: int
                n_jobs for UMAP.
            umap_init: str
                init for UMAP.
            densmap: boolean
                densmap for UMAP.
            min_cluster_size: int
                The parameter of HBDSACN from user.
            min_samples: int
                The parameter of HBDSACN from user.
            cluster_selection_method: string
                The parameter of HBDSACN from user.
            core_dist_n_jobs: int
                The parameter of HBDSACN from user.
            softclustering: bool
                The parameter of HBDSACN from user.
            weblog_unit: str
                The parameter for Weblogo.
            weblog_first_index: str
                The parameter for Weblogo.
        ---------

        Return
        ---------
            hdbscan-parameter-list: dict
                The parameters of HDBSCAN from user.
            processing_1: div object
                The information about data processing.
            processing_3: div object
                The information about data processing.
            upload_data: dict
                A Div style of 'upload_data' to hide the div object.
            para-list-div: dict
                A Div style of 'para-list-div' to hide the div.
            upload-div: dict
                A Div style of 'upload-div' to hide the div.
        """
        nonlocal input_validated

        dict_parameters = {
            "exp_len": exp_len,
            "n_neighbors": n_neighbors,
            "min_dist": min_dist,
            "random_state": random_state,
            "umap_jobs": umap_jobs,
            "umap_init": umap_init,
            "densmap": densmap,
            "min_cluster_size": min_cluster_size,
            "min_samples": min_samples,
            "cluster_selection_method": cluster_selection_method,
            "cluster_selection_epsilon": cluster_selection_epsilon,
            "hdbscan_jobs": hdbscan_jobs,
            "softclustering": softclustering,
            "weblogo_unit": weblogo_unit,
            "weblogo_first_index": weblogo_first_index,
            "weblogo_base_type": weblogo_base_type,
        }
        
        if input_validated == True:
            hide_div_style_1 = {'display': 'none'}
            hide_div_style_2 = {'display': 'none'}
            hide_div_style_3 = {'display': 'none'}
            if list_of_contents is not None:
                for c in list_of_contents:
                    return dict_parameters, [html.H3(time.asctime( time.localtime(time.time())))],html.H3('Processing ......',id='process'),hide_div_style_1,hide_div_style_2,hide_div_style_3
        else:
            return dash.no_update

    @app.callback(
                Output('processing_2', 'children'),
                Output('submit-data', 'style'),
                Output('cluster-data', 'data'),
                Output('processing_3','style'),
                Output('horizontal_line','style'),
                Input('hdbscan-parameter-list','data'),
            [State('upload_data', 'contents')],
                prevent_initial_call=True)

    def upload_file(parameter_list, list_of_contents):

        """
        Parameters
        ---------
            parameter_list: list
                The parameters of HDBSCAN from user.
            list_of_contents: list
                A content of upload file.

        ---------

        Returns
        ---------
            processing_2: div object
                The information about data processing.
            style: dict
                A style to hide 'submit-data' div object.
            parameters_dict: dict
                The results of clustering with HDBSCAN. 
            hide_info_style: dict
                A style to hide 'processing_2' div object.
            display_hr: dict
                A style to display horizontal line.
        """  
        nonlocal iMVP_finished

        hide_info_style  = {'display':'none'}
        display_hr = {'display':'inline-block'}

        if list_of_contents is not None:
            for c in list_of_contents:
                iMVP_out =  run_iMVP(c, parameter_list)
                if len(iMVP_out) == 2:
                    processing_2, style = iMVP_out
                    parameters_dict = None
                else:
                    processing_2, style, parameters_dict = iMVP_out
                    iMVP_finished = True
        return processing_2, style, parameters_dict, hide_info_style, display_hr


    @app.callback(
        Output('cluster_figure', 'figure'),
        Output('my-checklist','options'),
        Output('type', 'data'),
        Output('hidden_data','style'),
        Output('submit-button','style'),
        [Input('submit-data','n_clicks'),
        State('cluster-data', 'data'),
        Input("scatter-slider", "value"), 
        ],prevent_initial_call=True
    )

    def cluster_figure(n_clicks, cluster_data, markersize):
        """
        Parameters
        ---------
            n_clicks: int
                The number of clicks to trigger clustering.
            cluster_data: dict
                The results of clustering with HDBSCAN.
            markersize:
                The size of markers
        ---------

        Returns
        ---------
            cluster_figure:  graph object in plotly
                A graph to display the clusters of HBDSCAN.
            my-checklist: list
                The types of cluster that user choosed.
            type: list
                Types of clusters.
            hidden_data: dict
                A style to hide the div object.
            submit-button: button object
                A style to show the div of button object. 

        """
        dff = pd.DataFrame(cluster_data)
        df = dff.sort_values(by="Cluster", ascending=True)
        type = range(1,max(df['Cluster']) + 1)
        df['Cluster'] = df['Cluster'].astype(str)
        available_type =  list(map(str, type)) 
        df['customdata'] = df.index.values.tolist()
        options = [{'label': '{:>3}'.format(i), 'value':i } for i in available_type]
        fig = px.scatter(df, x="X", y="Y", color="Cluster", custom_data=["customdata"])
        fig.update_traces(marker_size=markersize) # , selector=dict(mode='markers')
        fig.update_layout(dragmode='lasso', hovermode=False, width=600, height=600)
        
        return fig, options,available_type,{'display':'inline-block', 'min-width': "1200px"},{'display':'none'} # 

    @app.callback(
        Output("my-checklist", "value"),
        Output("all-or-none", "value"),
        Output("select-data","data"),
        [Input("type", 'data'),
        Input("all-or-none", "value"),
        Input("my-checklist", 'value')],
    )

    def select_all_none(option,all_selected, my_selected):
        """
        Parameters
        ---------
            option: list
                Types of all clusters.
            all_selected: list
                When user choose all clusters. 
            my_selected: list 
                Types of clusters that user choosed.

        ---------

        Returns
        ---------
            my-checklist: list
                Types of clusters that user choosed, which show as checklist object.
            all-or-none: list
                Types of all clusters, which show as checklist object.
            select-data: list
                Types of clusters that user choosed, which store as dcc object.

        """
        ctx = callback_context
        input_id = ctx.triggered[0]["prop_id"].split(".")[0]
        
        if input_id == "my-checklist":
            all_selected = ["Select All"] if set(my_selected) == set(option) else []
        else:
            my_selected = option if all_selected else []
        if all_selected != []:
            select_data = option
        else:
            select_data = my_selected

        return my_selected, all_selected, select_data

    @app.callback(
        Output('weblogo','src'),
        Input('cluster_figure', 'selectedData'),
        Input('cluster-data', 'data'),
        Input('select-data','data'),
        Input('hdbscan-parameter-list','data'),
        prevent_initial_call=True
        )

    def draw_weblogo(selected_data,cluster_data, clusters_select, parameter_list):
        """
        Parameters
        ---------
            selected_data: dict
                Data selected with lasso or checklist
            cluster_data: dict
                The results of clustering with HDBSCAN. 
            clusters_select: list 
                Types of clusters selected from users
        ---------

        Return
        ---------
            weblogo: png
                Weblogo picture in png format.
        ---------
        """
        nonlocal output_path, iMVP_finished, first_time
        if iMVP_finished == True and first_time == True:
            # and os.path.isfile("{path}/selected_data.fa".format(path=output_path)) == False
            img_file = '{path}/weblogo.png'.format(path=output_path)
            encode_img = base64.b64encode(open(img_file,'rb').read())
            iMVP_finished = True
            first_time = False
            return 'data:image/png;base64,{}'.format(encode_img.decode())
        elif clusters_select == [] and selected_data is None:
            return dash.no_update
        else:
            df = pd.DataFrame(cluster_data)
            df['Cluster'] = df['Cluster'].astype(str)
            fa_index = []
            if selected_data is None:
                custom = []
                selected_data = {}
                for i in df[df['Cluster'].isin(clusters_select) ].index.values.tolist():
                    custom.append({'customdata':[i]})
                selected_data['points'] = custom
            
            for points in selected_data['points']: 
                fa_index.extend(points.get('customdata'))
            
            fasta_name = "{path}/selected_data.fa".format(path=output_path)

            base_type = parameter_list["weblogo_base_type"]

            with open(fasta_name, "w") as fasta_out:
                for idx, row in df.loc[fa_index].iterrows():
                    if base_type == "DNA":
                        seq_out = str(row["seq"]).upper().replace("U", "T")
                    elif base_type == "RNA":
                        seq_out = str(row["seq"]).upper().replace("T", "U")
                    else:
                        seq_out = str(row["seq"])
                    fasta_out.write(">{}\n{}\n".format(idx, seq_out))

            png = interactive_functions.draw_logo(fasta_name, parameter_list)
            image = Image.open(io.BytesIO(png))
            img_file = '{path}/weblogo.png'.format(path=output_path)
            image.save(img_file) 
            encode_img = base64.b64encode(open(img_file,'rb').read())
            
            return 'data:image/png;base64,{}'.format(encode_img.decode())

    @app.callback(
        Output("download-text", "data"),
        Input("btn-download-txt", "n_clicks"),
        prevent_initial_call=True)
    def download_fasta(n_clicks):
        """
        Parameters
        ---------
            n_clicks: int
                The number of clicks to trigger download file in fasta format.
        ---------

        Return
        ---------
            download-text: string   
                A fasta format file.
        ---------
        """
        nonlocal output_path
        with open('{path}/selected_data.fa'.format(path=output_path)) as f:
            contents = f.read()
        return dict(content=contents, filename="seleted_data.fa")

    @app.callback(
        Output("download-csv", "data"),
        Input("btn-download-csv", "n_clicks"),
        prevent_initial_call=True)
    def download_csv(n_clicks):
        """
        Parameters
        ---------
            n_clicks: int
                The number of clicks to trigger download CSV file.
        ---------

        Return
        ---------
            download-text: string   
                A fasta format file.
        ---------
        """
        nonlocal output_path
        with open('{path}/all_clusters.csv'.format(path=output_path)) as f:
            contents = f.read()
        return dict(content=contents, filename="all_clusters.csv")

    @app.callback(
        Output("download-png", "data"),   
        Input("btn-download-png", "n_clicks"),
        prevent_initial_call=True
    )
    def download_weblogo(n_clicks):
        nonlocal output_path
        """
        Parameters
        ---------
            n_clicks: int 
                The number of clicks to trigger download weblogo picture.
        ---------

        Return
        ---------
            download-png: png
                A file in png format of weblogo.
        ---------
        """
        return dash.dcc.send_file("{path}/weblogo.png".format(path=output_path))

    app.layout = html.Div([



        html.Div([
            html.Div([
                html.H1(
                        "iMVP Motif Viewer",
                        style = {'textAlign':'center'}), # , 'margin-left':'20%'

                html.H3(
                        "Version: 0.2.3; Contributed by Jing Yao, Jianheng Liu @ Zhang Lab (SYSU).",
                        style = {'textAlign':'center'}), # , 'margin-left': '10%'
                html.H3("Documents: https://readthedocs.org/iMVP/", style = {'textAlign':'center'}),

                html.Div([
                    # html.Br(),
                    # html.Br(),
                    html.Div("Tips #1: To go back to the parameters page, please refresh the page."),
                    html.Div("Tips #2: Use Ctrl+C in command lines to terminate the backend."),
                ], style={"horizonal-align":"center", 'text-align':'center'}),
            ], style={"width":"1000px"}),

            html.Div([
                html.Div([
                    html.Div([
                        html.Div([
                            html.Div([
                                html.H4('I. Upload data'),
                                dcc.Upload(
                                    id = 'upload_data',
                                    multiple=True,
                                    style={"line-height": "300px"}  # , "min-width": "100%
                                ), 

                            ], className = "upload",id = 'upload-div', style={'display':'inline-block', "width": "600px"},  # "line-height: 20%;"
                            ),

                        html.Div([
                                html.Br(),
                                html.Br(),
                                html.Div("If you have confirmed your input and parameters, click the button to run."),
                                html.Br(),
                                html.Button(id='submit-button-state', n_clicks=0, children='Submit'),
                            ],
                            className="input", style={"horizonal-align":"center", 'text-align':'center'},
                        ),

                    ], className="two columns", style={'display':'inline-block', "width": "600px", "vertical-align":"top", "margin-right":"5%", "margin-left":"2.5%"}
                    ),

                    # html.Div([

                    #     ], className="three columns", style={'display':'inline-block'}
                    # ),

                    html.Div([
                        html.Div([
                            html.H4('II. Quality control'),
                            html.Div([
                                "1. Expected lengths of the sequences =",
                                dcc.Input(id='exp_len', type='number', value='21', min='0'),
                            ]),
                            ], style={'display':'inline-block', "vertical-align":"top"}),

                            html.H4('III. UMAP parameters'),
                            html.Div([
                                "1. n_neighbors =",
                                dcc.Input(id='n_neighbors', type='number', value='20', min='0'),
                                ]),
                            html.Div([
                                "2. min_dist =", 
                                dcc.Input(id='min_dist', type='number', value='0.01', step='any', min='0', max='1'), 
                                ]),
                            html.Div([
                                "3. random_state =", 
                                dcc.Input(id='random_state', type='number', value='42', min='-2147483648', max='2147483647'), 
                                ]),
                            html.Div([
                                "4. jobs =", 
                                dcc.Input(id='umap_jobs', type='number', value='6'), 
                                ]),
                            html.Div([
                                html.Div(["5. init ="], className="two columns", style={"display": "inline-block"}), # style={"display": "inline-block"}
                                html.Div([dcc.RadioItems(['random', 'spectral'], 'random', id='umap_init')], className="two columns", style={"display": "inline-block"}),        
                            ], className="row"# style={"width":"30%"}
                            ),
                            html.Div([
                                html.Div(["6. DensMAP ="], className="two columns", style={"display": "inline-block"}), # style={"display": "inline-block"}
                                html.Div([dcc.RadioItems(['True', 'False'], 'False', id='densmap')], className="two columns", style={"display": "inline-block"}),        
                            ], className="row"# style={"width":"30%"}
                            ),

                            html.H4('IV. HBDSCAN parameters'),
                            html.Div([
                                "1. min_cluster_size =",
                                dcc.Input(id='min_cluster_size', type='number', value='100'),
                                ]),
                            html.Div([
                                "2. min_samples =", 
                                dcc.Input(id='min_samples', type='number', value='100'),
                                ]),
                            html.Div([
                                "3. cluster_selection_epsilon =", 
                                dcc.Input(id='cluster_selection_epsilon', type='number', step="any", value='0.0'),
                                ]),

                            html.Div([
                                html.Div(["4. cluster_selection_method ="], className="two columns", style={"display": "inline-block"}), # style={"display": "inline-block"}
                                html.Div([dcc.RadioItems(['eom', 'leaf'], 'eom', id='cluster_selection_method')], className="two columns", style={"display": "inline-block"}),        
                            ], className="row"# style={"width":"30%"}
                            ),

                            html.Div([
                                html.Div(["5. soft clustering ="], className="two columns", style={"display": "inline-block"}), # style={"display": "inline-block"}
                                html.Div([dcc.RadioItems(['True', 'False'], 'True', id='softclustering')], className="two columns", style={"display": "inline-block"}),        
                            ], className="row"# style={"width":"30%"}
                            ),
                            html.Div([
                                "6. jobs =", 
                                dcc.Input(id='hdbscan_jobs', type='number', value='6'),
                                ]),

                            html.H4('V. Weblogo'),
                            html.Div([
                                html.Div(["1. Unit ="], className="two columns", style={"display": "inline-block"}), # style={"display": "inline-block"}
                                html.Div([dcc.RadioItems(['probability', 'bits'], 'probability', id='weblogo_unit')], className="two columns", style={"display": "inline-block"}),        
                            ], className="row"# style={"width":"30%"}
                            ),
                            html.Div([
                                html.Div(["2. Base type (LOGO and FASTA output) ="], className="two columns", style={"display": "inline-block"}), # style={"display": "inline-block"}
                                html.Div([dcc.RadioItems(['As input', 'DNA', "RNA"], 'As input', id='weblogo_base_type')], className="two columns", style={"display": "block"}),  # inline-    
                            ], className="row"# style={"width":"30%"}
                            ),

                            html.Div(["3. First index =",
                                        dcc.Input(id='weblogo_first_index', type='number', value='-10'),
                                ]), 
                                
                        ], className="two columns", style={'display':'inline-block'} 
                        ),

                    ], style={'display':'inline-block', "margin-left":"0%", "width":"100%"}, # , "width":"40%"
                    ),
                ], className="row", style={'display':'inline-block', "width":"100%"}
                ),

                        
            ], id="para-list-div", style={"width":"auto"})

        ], className="section1",id = 'section1', style={"width":"1600px"}),

        html.Hr(id = "horizontal_line",style={'display':'none'}),

        html.Div([
            html.Div([
                html.Div([
                    html.Div(
                        dcc.Graph(id = 'cluster_figure'),
                        style={'display': 'inline-block'} # 'width': '40%',
                    ),
                    html.Div(
                        html.H4("Marker size:"),
                        style={"margin-left": "10%"}
                    ),
                    html.Div(
                            dcc.Slider(id='scatter-slider', value=10, min=1, max=30,),
                            style={"margin-left": "10%", "margin-right": "10%"}
                    ), 

                ], className="two columns", style={"min-width":"600px", "display": "inline-block", "max-width": "40%", "margin-right": "5%"} # , "margin-right": "%"
                ),

                html.Div([
                    html.Div([
                        html.Div(
                            html.Img(id = "weblogo", style={'max-width':"600px",'display': 'inline-block'}), 
                            style={'display': 'inline-block'}   # , 'verticalAlign':'top'
                            ), 
                            # style={'width':"200px"}), # 'width': '50%','display': 'inline-block', 'position':'relative','bottom':'150px'

                    html.Div([
                        html.Div([
                            html.Div(html.Button("Download FASTA", id = "btn-download-txt"), # , style = {'width': '99%'}
                            ),
                            dcc.Download(id = "download-text")
                        ],
                        style={'display':'inline-block', "margin-left": "10%", "margin-right":"5%"}, 
                        className='three columns'), 

                        html.Div([
                            html.Div(html.Button("Download CSV", id="btn-download-csv"), # , style={'width': '99%'}
                            ),
                            dcc.Download(id = "download-csv"),
                        ],
                        style={'display':'inline-block', "margin-left":"5%", "margin-right":"5%"}, 
                        className='three columns'),

                        html.Div([
                            html.Div(html.Button("Download LOGO", id="btn-download-png"), # , style={'width': '99%'}
                            ),
                            dcc.Download(id = "download-png"),
                        ],
                        style={'display':'inline-block', "margin-left":"5%", "margin-right": "10%"}, 
                        className='three columns'),

                    ], className="row", style={"horizonal-align": "middle", "margin-left": "5%", "margin-top": "5%", "margin-right": "5%", "margin-bottom": "5%"} # 'display': 'inline-block', 
                    ),

                    html.Br(),
                    html.Br(),

                    html.Div([
                        html.Div(
                            html.H4("Select clusters:", style={'display': 'inline-block'}),
                        ),
                        
                        html.Div([
                            dcc.Checklist(
                                id="all-or-none",
                                options=[{"label": "Select All", "value": "Select All"}],
                                value=[],
                                # labelStyle={"display": "inline-block", "position": "relative", "vertical-align":"middle"},
                            ),
                        ],  # style={'display': 'inline-block'}
                        ),

                        html.Div([
                            dcc.Checklist(
                                id="my-checklist",
                                #options=[{"label": x, "value": x} for x in option],
                                value=["1"],
                                # labelStyle={"display": "inline-block", "position": "relative", "vertical-align":"middle"}, # "text-": "relative"
                            ),
                        ], # style={'display': 'inline-block'}
                        ),
                    ], style={'display': 'inline-block'}),

                    ], style={'display': 'inline-block'}
                    ),

                ], className="two columns", style={"margin-top":"5%", "max-width": "40%", "display": "inline-block",'vertical-align': 'top'} 
                ),
            ], className="row", style={"display": "inline-block", "min-width":"1000px"} 
            ),
            html.Br(),
            html.Br(),
            html.Br(),

        ], id = "hidden_data", style={'display':'none', 'min-width': '1000px'},
        ),

        

        # html.Div(
        #     dcc.Markdown('''
        #         *Usage:*
        #             The software encodes the data using one-hot encoding and the dimensionality reduction using **UMAP**. Then, the matrix is clustered using **HDBSCAN** to get all the clusters from the fasta file.Firstly, upload the fasta data of the same length. Then, input the parameters for the clustering with HDBSCAN. After submitting the data, it will take a few minutes for the background to process the data.Finally, you can select clusters by tick checklist, or use a lasso to circle the parts of interest on the cluster plot. That part of data would display by weblogo of base enrichment.
        #         ''')
        # )

        html.Div(id = "processing_1"),
        html.Div(id = "processing_2"),
        html.Div(id = "processing_3"),
        html.Div(
            html.Button(
                    id = 'submit-data',
                    n_clicks=0,
                    children='Draw the Figures',
                    style = {'display':'none'}),
            id = 'submit-button'),

        dcc.Store(id = "cluster-data"),
        dcc.Store(id = "hdbscan-parameter-list"),
        dcc.Store(id = "select-data"),
        dcc.Store(id = "type")
    
    ], style={"margin-left":"0%", "width":"1000px"})

    return app

if __name__ == "__main__":
    pass


from setuptools import setup

setup(
    name = 'iMVP-utils',
    author = 'Jianheng Liu, Jing Yao',
    author_email='jhfoxliu@gmail.com',
    url="https://github.com/jhfoxliu/iMVP",
    project_urls={
        "Bug Tracker": "https://github.com/jhfoxliu/issues",
    },
    description = 'Utils for interactive epitranscriptomic Motif Visualization and Sub-type Partitioning (iMVP).',
    version = '0.2.3',
    packages = ['iMVP_utils'],
    license = "MIT",
    scripts=["iMVP_viewer.py"],
    python_requires=">=3.6",
    install_requires = [
        "numpy>=1.20.0",
        "pandas>=1.3.4",
        "scipy>=1.5.1",
        "hdbscan>=0.8.27",
        "umap-learn>=0.5.2",
        "matplotlib>=3.2.2",
        "seaborn>=0.10.1",
        "scikit-learn>=0.23.1",
        "biopython>=1.77",
        "dash>=2.2.0",
        "dash-bio>=0.9.0",
        "imageio>=2.13.5",
        "weblogo>=3.7.0",
        "opencv-python>=4.5.5",
    ],
    package_data={"iMVP_utils": ["assets/*.css"]},

)


'''iMVP plots

'''

def show_logos_cols(prefix, names=None, cols=3, figsize=(8,8),
                    auto_size=True, auto_width=4, auto_height=1.5,
                    savefig_name=None, dpi=300):
    """This function is used for plot a series of motif logos in PNG format.
   
    Parameters
    ----------
    prefix: str
        The name of output path, required. This function will scan all PNG files in this path.
    names: tuple
        If given, only plot the given file names. Default=None
    cols: int
        The number of columns. Default=3
    figsize: tuple
        The figsize parameter for matplotlib.pyplot.subpolots()
    auto_size: boolean
        If True, ignore figsze and compute the width and height automatically.
    auto_width: float
        The width factor used for auto_size.
    auto_height: float
        The height factor used for auto_size.
    savefig_name: str
        The plot to save, should end with .pdf or .png or ect. If None, figure will not be drawn.
    dpi: int
        The dpi value for the figure.
    Returns
    -------
    matplotlib.axes
    """
    import os
    import matplotlib.pyplot as plt

    file_list = []
    for img in os.listdir(prefix):
        if img.endswith(".png") == False:
            continue
        if names is not None and fn not in names:
            continue
        if os.path.getsize(prefix+"/"+img) == 0:  # empty file, skipped
            continue
        file_list.append(img)
    
    file_list_format = []
    for i in file_list:
        id = int(i.replace("cluster_", "").replace(".png", ""))
        file_list_format.append((i, id))
    file_list_format = sorted(file_list_format, key=lambda x:x[1]) 
    
    if len(file_list_format) % cols == 0:
        rows = len(file_list_format) // cols
    else:
        rows = len(file_list_format) // cols + 1
    if auto_size == False:
        figsize = figsize
    else:
        width = auto_width * cols
        height = auto_height * rows
        figsize = (width, height)
    
    if len(file_list_format) > 1:
        fig, axes = plt.subplots(rows, cols, figsize=figsize)
        for ax, image in zip(*[axes.reshape(-1), file_list_format]):
            fn, id = image
            img = plt.imread(prefix+"/"+fn)
            _ = ax.imshow(img)
            ax.set_title("cluster #{}".format(id))
        for ax in axes.reshape(-1):
            ax.axis("off")
        if savefig_name is not None:
            plt.savefig(savefig_name, dpi=dpi)
        plt.tight_layout()
        return axes
    else:
        fig, ax = plt.subplots(figsize=figsize)
        fn, id = file_list_format[0]
        img = plt.imread(prefix+"/"+fn)
        _ = ax.imshow(img)
        ax.set_title("cluster #{}".format(id),size=16)
        ax.axis("off")
        plt.tight_layout()
        if savefig_name is not None:
            plt.savefig(savefig_name, dpi=dpi)
        return ax
    
def draw_2D_hist(df, vmax=0.05, cmin = None, density = True, 
                 xlim = None, ylim = None,
                 bins = [600, 600]):
    """This function is used for draw a 2D histogram.
   
    Parameters
    ----------
    df: pd.DataFrame
        A DataFrame containing the columns X and Y.
    vmax: float
        The vmax parameter for hist2d. Default=0.05.
    cmin: float
        The cmin parameter for hist2d. Default=None.
    density: boolean
        If draw density histogram. Default=True.
    xlim: tuple
        xlim for hist2d. Default=None.
    ylim: tuple
        ylim for hist2d. Default=None.
    bins: tuple
        Bin numbers for hist2d. Default=[600,600]
    Returns
    -------
    hist2d: np.array
        A 2D array representing the values of the histogram. Please note that this array has been 90-degree rotated to fit the real X-Y and hence can be drawn with plt.imshow() directly.
    edgesX: np.array
        The X edges.
    edgesY: np.array
        The Y edges
    """
    import matplotlib.pyplot  as plt
    import numpy as np
    fig, ax = plt.subplots(figsize=[6,6])
    ax.set_aspect('equal', 'box')
    hist2d, edgesX, edgesY, ax_hist = ax.hist2d(df["X"], df["Y"], range=[xlim, ylim], vmax=vmax, bins=bins, cmin=cmin, density=density)
    
    # small trick here, the x,y row/col is different in figure and numpy, let's make it the same as we view
    hist2d = np.rot90(hist2d)
    
    ax.set_xticks([])
    ax.set_yticks([])
    plt.tight_layout()
    return hist2d, edgesX, edgesY

if __name__ == "__main__":
    pass